
// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA, Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_basic_offensive_kick.h"

#include "body_kick_to_corner.h"
#include "pass_course_table.h"
#include <rcsc-ext/player/world_model_ext.h>

#include <rcsc/action/body_advance_ball.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/body_hold_ball.h>
#include <rcsc/action/body_shoot.h>
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_stop_ball.h>
#if 0
#include <rcsc/action/body_clear_ball.h>
#endif
#include <rcsc/action/bhv_scan_field.h>
#include <rcsc/action/neck_scan_field.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>
#include <rcsc/action/neck_turn_to_ball_or_scan.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc-ext/action/bhv_find_player.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/say_message_builder.h>

#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/geom/sector_2d.h>
#include <rcsc/math_util.h>

#include <set>
#include <deque>

#include "simulator_util.h"
#include "simulator.h"
#include "body_advance_ball2008.h"
#include "body_clear_ball2008.h"
#include "body_shoot_test.h"
#include "body_force_shoot.h"

#include "team_switch.h"

#if defined(USE_SMART_KICK)
#include "contrib/body_smart_kick.h"
#endif


#define ALLOW_SELF_PASS 0

#define USE_FORCE_SHOOT


static
std::pair< bool, double >
evaluateSituation( const SimulatorUtil::SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent );

static
double
s_get_ball_speed_for_pass( double distance );

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_BasicOffensiveKick::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        "%s:%d: Bhv_BasicOffensiveKick"
                        ,__FILE__, __LINE__ );

    const rcsc::WorldModel & wm = agent->world();
    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
    const rcsc::PlayerObject * nearest_opp
        = ( opps.empty()
            ? static_cast< rcsc::PlayerObject * >( 0 )
            : opps.front() );
    const double nearest_opp_dist = ( nearest_opp
                                      ? nearest_opp->distFromSelf()
                                      : 1000.0 );
    const rcsc::Vector2D nearest_opp_pos = ( nearest_opp
                                             ? nearest_opp->pos()
                                             : rcsc::Vector2D( -1000.0, 0.0 ) );

    if ( rcsc::Body_Shoot().execute( agent ) )
    {
        return true;
    }

    if ( (wm.self().pos() - param.theirTeamGoalPos()).r() < 20.0 )
    {
        if ( Body_Shoot_Test().execute( agent ) )
        {
            return true;
        }
    }

#if defined(USE_FORCE_SHOOT)
    if ( (wm.self().pos() - rcsc_ext::ServerParam_opponentTeamNearGoalPostPos
                                                           ( wm.self().pos() ))
         .r() < 5.0 )
    {
        return Body_ForceShoot().execute( agent );
    }
#endif


#if 0
    //
    // clear ball
    //
    const rcsc::Rect2D clear_area( param.ourTeamGoalLineX(),
                                   - param.penaltyAreaHalfWidth(),
                                   param.penaltyAreaLength() + 5.0,
                                   param.penaltyAreaWidth() );

# if 0
    if ( clear_area.contains( wm.self().pos() )
         || wm.existKickableOpponent() )
    {
        if ( Body_ClearBall2008().execute( agent ) )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
            agent->debugClient().addMessage( "clear ball" );
            rcsc::dlog.addText( rcsc::Logger::TEAM, "clear ball" );
            return true;
        }
    }
# endif
#endif


    const PassCourseTable & table = PassCourseTable::instance( wm );

    const PassCourseTable::CourseSelection first_pass = table.get_first_selection( agent );
    const PassCourseTable::CourseSelection pass_result = table.get_result( agent );

#if 0
    if ( wm.self().pos().x < 30.0 )
    {
        if ( Body_AdvanceBall2008().execute( agent ) )
        {
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            agent->debugClient().addMessage( "advance ball" );
            rcsc::dlog.addText( rcsc::Logger::TEAM, "advance ball" );
            return true;
        }
    }
#endif

    boost::shared_ptr< SimulatorUtil::SimulatorEstimatedSituation > best;
    {
        //
        // create plans
        //
        SimulatorUtil::KickPlanCont kick_plans;

        double search_width = 90.0;
        if ( wm.self().pos().x >= 40.0 )
        {
            search_width = 135.0;
        }

        for ( double th = -search_width; th < +search_width + rcsc::EPS; th += 10.0 )
        {
            const double vel = 2.5;

            const rcsc::Vector2D target( wm.self().pos()
                                         + rcsc::Vector2D::polar2vector
                                                           ( 10.0, th ) );
            const rcsc::Vector2D result_pos = wm.self().pos();
            const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                              ( vel, (target - wm.ball().pos()).th() );

            int kick_step = 2;

            kick_plans.push_back( SimulatorUtil::KickPlan
                                  ( boost::shared_ptr< rcsc::BodyAction >
#if defined(USE_SMART_KICK)
                                    ( new rcsc::Body_SmartKick
                                      ( target,
                                        param.ballSpeedMax(),
                                        param.ballSpeedMax() * 0.96,
                                        3 ) ),
#else
                                    ( new rcsc::Body_KickMultiStep
                                      ( target,
                                        param.ballSpeedMax(),
                                        false ) ),
#endif
                                    result_pos,
                                    result_vel,
                                    kick_step ) );
        }

#if 0
        for ( double th = -search_width; th < +search_width + rcsc::EPS; th += 10.0 )
        {
            const rcsc::Vector2D target( wm.self().pos()
                                         + rcsc::Vector2D::polar2vector
                                                           ( 10.0, th ) );
            double target_ball_speed
                   = rcsc::Body_KickOneStep::get_max_possible_vel
                     ( (target - wm.ball().pos()).th(),
                       wm.self().kickRate(),
                       wm.ball().vel() ).r();

            const rcsc::Vector2D result_pos = wm.ball().pos();
            const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                              ( target_ball_speed,
                                                (target
                                                 - wm.ball().pos()).th() );
                                           
            int kick_step = 1;

            kick_plans.push_back( SimulatorUtil::KickPlan
                                  ( boost::shared_ptr< rcsc::BodyAction >
                                    ( new rcsc::Body_KickOneStep
                                      ( target, target_ball_speed ) ),
                                    result_pos,
                                    result_vel,
                                    kick_step ) );
        }
#endif

#if 1
        //
        // between defenders
        //
        {
            std::deque<double> y_list;

            rcsc::AbstractPlayerCont defenders = wm.getPlayerCont
                ( new rcsc::AndPlayerPredicate
                  ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                    new rcsc::XCoordinateForwardPlayerPredicate
                    ( wm.offsideLineX() - 3.0 ),
                    new rcsc::XCoordinateBackwardPlayerPredicate
                    ( wm.offsideLineX() + 3.0 ) ) );

            std::set< double > defenders_y;

            const rcsc::AbstractPlayerCont::const_iterator d_end = defenders.end();
            for ( rcsc::AbstractPlayerCont::const_iterator it = defenders.begin();
                  it != d_end;
                  ++ it )
            {
                defenders_y.insert( (**it).pos().y );
            }

            const std::set< double >::const_iterator y_end = defenders_y.end();
            for ( std::set< double >::const_iterator it = defenders_y.begin();
                  it != y_end;
                  ++ it )
            {
                std::set< double >::const_iterator next = it;
                ++ next;

                if ( next == y_end )
                {
                    break;
                }

                y_list.push_back( ((*it) + (*next)) / 2.0 );
            }

            if ( ! y_list.empty() )
            {
                y_list.push_front
                    ( (- rcsc::ServerParam::i().pitchHalfWidth()
                       + *(y_list.begin())) / 2.0 );

                y_list.push_front
                    ( (+ rcsc::ServerParam::i().pitchHalfWidth()
                       + *(y_list.rbegin())) / 2.0 );
            }

            y_list.push_front
                   ( - (rcsc::ServerParam::i().pitchHalfWidth() - 5.0 ));
            y_list.push_back
                   ( + (rcsc::ServerParam::i().pitchHalfWidth() - 5.0 ));

            for ( size_t i = 0; i < y_list.size(); ++ i )
            {
                const rcsc::Vector2D target( wm.offsideLineX(), y_list[i] );

                double target_ball_speed
                    = rcsc::Body_KickOneStep::get_max_possible_vel
                    ( (target - wm.ball().pos()).th(),
                      wm.self().kickRate(),
                      wm.ball().vel() ).r();

                const rcsc::Vector2D result_pos = wm.ball().pos();
                const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                    ( target_ball_speed,
                      (target
                       - wm.ball().pos()).th() );
                                           
                int kick_step = 1;

                kick_plans.push_back( SimulatorUtil::KickPlan
                                      ( boost::shared_ptr< rcsc::BodyAction >
                                        ( new rcsc::Body_KickOneStep
                                          ( target, target_ball_speed ) ),
                                        result_pos,
                                        result_vel,
                                        kick_step ) );
            }

            for ( size_t i = 0; i < y_list.size(); ++ i )
            {
                const double vel = 2.5;

                const rcsc::Vector2D target( wm.offsideLineX(), y_list[i] );

                const rcsc::Vector2D result_pos = wm.self().pos();
                const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                              ( vel, (target - wm.ball().pos()).th() );

                int kick_step = 2;

                kick_plans.push_back( SimulatorUtil::KickPlan
                                      ( boost::shared_ptr< rcsc::BodyAction >
#if defined(USE_SMART_KICK)
                                        ( new rcsc::Body_SmartKick
                                          ( target,
                                            param.ballSpeedMax(),
                                            param.ballSpeedMax() * 0.96,
                                            3 ) ),
#else
                                        ( new rcsc::Body_KickMultiStep
                                          ( target,
                                            2.5,
                                            true ) ),
#endif
                                        result_pos,
                                        result_vel,
                                        kick_step ) );
            }
        }
#endif

#if 1
        //
        // between forwards
        //
        {
            std::deque<double> y_list;

            rcsc::AbstractPlayerCont forwards = wm.getPlayerCont
                ( new rcsc::AndPlayerPredicate
                  ( new rcsc::TeammatePlayerPredicate( wm ),
                    new rcsc::XCoordinateForwardPlayerPredicate
                    ( wm.offsideLineX() - 10.0 ),
                    new rcsc::XCoordinateBackwardPlayerPredicate
                    ( wm.offsideLineX() ) ) );

            const rcsc::AbstractPlayerCont::const_iterator a_end = forwards.end();
            for ( rcsc::AbstractPlayerCont::const_iterator it = forwards.begin();
                  it != a_end;
                  ++ it )
            {
                y_list.push_back( (**it).pos().y );
            }

            for ( size_t i = 0; i < y_list.size(); ++ i )
            {
                const rcsc::Vector2D target( wm.offsideLineX() + 10.0,
                                             y_list[i] );

                double target_ball_speed
                    = rcsc::Body_KickOneStep::get_max_possible_vel
                    ( (target - wm.ball().pos()).th(),
                      wm.self().kickRate(),
                      wm.ball().vel() ).r();

                const rcsc::Vector2D result_pos = wm.ball().pos();
                const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                    ( target_ball_speed,
                      (target
                       - wm.ball().pos()).th() );
                                           
                int kick_step = 1;

                kick_plans.push_back( SimulatorUtil::KickPlan
                                      ( boost::shared_ptr< rcsc::BodyAction >
                                        ( new rcsc::Body_KickOneStep
                                          ( target, target_ball_speed ) ),
                                        result_pos,
                                        result_vel,
                                        kick_step ) );
            }

            for ( size_t i = 0; i < y_list.size(); ++ i )
            {
                const double vel = 2.5;

                const rcsc::Vector2D target( wm.offsideLineX() + 10.0,
                                             y_list[i] );

                const rcsc::Vector2D result_pos = wm.self().pos();
                const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                              ( vel, (target - wm.ball().pos()).th() );

                int kick_step = 2;

                kick_plans.push_back( SimulatorUtil::KickPlan
                                      ( boost::shared_ptr< rcsc::BodyAction >
#if defined(USE_SMART_KICK)
                                        ( new rcsc::Body_SmartKick
                                          ( target,
                                            param.ballSpeedMax(),
                                            param.ballSpeedMax() * 0.96,
                                            3 ) ),
#else
                                        ( new rcsc::Body_KickMultiStep
                                          ( target,
                                            2.5,
                                            true ) ),
#endif
                                        result_pos,
                                        result_vel,
                                        kick_step ) );
            }
        }
#endif


#if 0
        int teammate_penalty_step = 3;
        int opponent_penalty_step = 1;
#elif 0
        // OK
        int teammate_penalty_step = 3;
        int opponent_penalty_step = 2;
#elif 0
        int teammate_penalty_step = 3;
        int opponent_penalty_step = 3;
#elif 0
        int teammate_penalty_step = 3;
        int opponent_penalty_step = 2;

        if ( wm.self().pos().x >= wm.offsideLineX() - 10.0 )
        {
            teammate_penalty_step = 3;
            opponent_penalty_step = 3;
        }
#elif 0
        int teammate_penalty_step = 1;
        int opponent_penalty_step = 2;
#elif 0
        // OK?
        int teammate_penalty_step = 1;
        int opponent_penalty_step = 3;
#elif 0
        int teammate_penalty_step = 1;
        int opponent_penalty_step = 4;
#else
        int teammate_penalty_step = 2;
        int opponent_penalty_step = 3;
#endif

#if 0
        if ( wm.ball().pos().x >= 40.0 )
        {
            teammate_penalty_step = 2;
            opponent_penalty_step = 2;
        }
        else if ( wm.self().pos().x >= wm.offsideLineX() - 10.0 )
        {
            opponent_penalty_step = 2;
        }
#elif 0
        if ( wm.ball().pos().x >= 40.0 )
        {
            teammate_penalty_step = 1;
            opponent_penalty_step = 5;
        }
#elif 0
        if ( wm.ball().pos().x >= 40.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 4;
        }
        else if ( wm.self().pos().x >= wm.offsideLineX() - 10.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 3;
        }
#elif 0
        // OK
        if ( wm.ball().pos().x >= 40.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 5;
        }
        else if ( wm.self().pos().x >= wm.offsideLineX() - 10.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 3;
        }
#elif 1
        if ( wm.ball().pos().x >= 40.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 3;
        }
        else if ( wm.self().pos().x >= wm.offsideLineX() - 10.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 3;
        }
#else
        if ( wm.ball().pos().x >= 45.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 5;
        }
        else if ( wm.ball().pos().x >= 40.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 3;
        }
        else if ( wm.self().pos().x >= wm.offsideLineX() - 10.0 )
        {
            teammate_penalty_step = 0;
            opponent_penalty_step = 3;
        }
#endif

        //
        // get the best plan
        //
        best = SimulatorUtil::getBestChoice
                              ( evaluateSituation,
                                kick_plans,
                                wm.getPlayerCont
                                ( new rcsc::CoordinateAccuratePlayerPredicate
                                                                       ( 20 ) ),
                                agent,
                                true, 10,
                                true, teammate_penalty_step,
                                true, opponent_penalty_step );

        //
        // execute
        //
        if ( best )
        {
            const SimulatorUtil::KickPlan & plan = best->plan;

            if ( plan.getAction()->execute( agent ) )
            {
                int pass_receiver = best->first_access_teammate;

                long our_team_access_step = std::min( best->teammate_access_step,
                                                      best->self_access_step );

                rcsc::Vector2D receive_point = rcsc::inertia_n_step_point
	                                        ( plan.getKickResultPos(),
                                                  plan.getKickResultVel(),
                                                  our_team_access_step,
                                                  param.ballDecay() );

                agent->setNeckAction( new rcsc::Neck_TurnToPoint
                                      ( wm.teammate( pass_receiver )
                                        ->inertiaPoint( 1 ) ) );

                agent->debugClient().addLine( wm.ball().pos(),
                                              receive_point );

                rcsc::Vector2D target_buf = receive_point - agent->world().self().pos();
                target_buf.setLength( 1.0 );

                agent->addSayMessage( new rcsc::PassMessage
                                      ( pass_receiver,
                                        receive_point + target_buf,
                                        agent->effector().queuedNextBallPos(),
                                        agent->effector().queuedNextBallVel() ) );
                return true;
            }
        }
    }


    const rcsc::Vector2D goal_pos = param.theirTeamGoalPos();
    agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );

    switch( first_pass.type )
    {
    case PassCourseTable::CourseSelection::Shoot:
        if ( Body_ForceShoot().execute( agent ) )
        {
            return true;
        }
        break;

    case PassCourseTable::CourseSelection::Push:
      {
        rcsc::Vector2D dribble_target;

#if 0
        if ( wm.self().pos().x > + 35.0 )
#else
        if ( wm.self().pos().x > + 20.0 )
#endif
        {
#if 0
            dribble_target = rcsc_ext::ServerParam_opponentTeamFarGoalPostPos
                                                           ( wm.self().pos() );
#else
            dribble_target = rcsc_ext::ServerParam_opponentTeamNearGoalPostPos
                                                           ( wm.self().pos() );
#endif
        }
        else
        {
            dribble_target = wm.self().pos() + rcsc::Vector2D( 10.0, 0.0 );
        }

#if 1
        return rcsc::Body_Dribble( dribble_target,
                                   1.0,
                                   param.maxPower(),
                                   3 ).execute( agent );
#else
        return rcsc::Body_Dribble( dribble_target,
                                   1.0,
                                   param.maxPower(),
                                   10 ).execute( agent );
#endif
        break;
      }

    case PassCourseTable::CourseSelection::FindPassReceiver:
    case PassCourseTable::CourseSelection::Pass:
        {
            const rcsc::AbstractPlayerObject * target_player = wm.teammate( first_pass.player_number );

            if ( nearest_opp_dist > 4.0
                 && first_pass.type == PassCourseTable::CourseSelection::FindPassReceiver )
            {
#if 0
                if ( ! agent->effector().queuedNextBallKickable()
                     && wm.ball().vel().r() >= 0.1 )
#else
                if ( (wm.ball().inertiaPoint( 1 )
                      - wm.self().inertiaPoint( 1 )).r()
                     > wm.self().kickableArea() - 0.2 )
#endif
                {
                    if ( rcsc::Body_StopBall().execute( agent ) )
                    {
                        agent-> setNeckAction( new rcsc::Neck_TurnToPoint
                                               ( target_player->pos() ) );

                        agent->debugClient().addMessage( "stop ball" );
                        agent->debugClient().setTarget( target_player->unum() );

                        rcsc::dlog.addText( rcsc::Logger::TEAM,
                                            "stop ball" );

                        return true;
                    }
                }

                if ( rcsc_ext::Bhv_FindPlayer( target_player )
                     .execute( agent ) )
                {
                    agent-> setNeckAction( new rcsc::Neck_TurnToPoint
                                           ( target_player->pos() ) );

                    agent->debugClient().addMessage( "find" );
                    agent->debugClient().setTarget( target_player->unum() );

                    rcsc::dlog.addText( rcsc::Logger::TEAM,
                                        "find" );

                    return true;
                }
            }

#if 1
            table.setDebugPassRoute( agent );
#endif

            long max_diff = 0;
            double max_a = -1;
            rcsc::Vector2D target_point = target_player->inertiaPoint( 1 );

            double target_ball_speed = s_get_ball_speed_for_pass
                                       ( (target_point
                                          - wm.self().pos()).r() );

#if 0
            Simulator sim;
            rcsc::AbstractPlayerCont players = wm.getPlayerCont
                ( new rcsc::CoordinateAccuratePlayerPredicate( 20 ) );

            long self_access_step;
            long teammate_access_step;
            long opponent_access_step;
            long unknown_access_step;
            int first_access_teammate;
            int first_access_opponent;

            for ( double a = -180.0; a < 180; a += 45.0 )
            {
                const rcsc::Vector2D target( target_player->pos()
                                             + rcsc::Vector2D::polar2vector
                                               ( 5.0, a ) );

                rcsc::dlog.addText( rcsc::Logger::ACTION,
                                    "a = %lf", a );
                rcsc::dlog.addText( rcsc::Logger::ACTION,
                                    "target = (%lf,%lf)",
                                    target.x, target.y );

#if 0
                agent->debugClient().addCircle( target, 0.5 );
#endif

                const rcsc::Vector2D result_pos = wm.self().pos();
                const rcsc::Vector2D result_vel
                    = rcsc::Vector2D::polar2vector
                                      ( target_ball_speed,
                                        (target - wm.ball().pos()).th() );

                int kick_step = 2;

                sim.simulate( wm,
                              players,
                              result_pos,
                              result_vel,
                              kick_step,
                              kick_step + 20, // self penalty steps
                              3, // teammate penalty steps
                              0, // opponent or unknown penalty steps
                              &self_access_step,
                              &teammate_access_step,
                              &opponent_access_step,
                              &unknown_access_step,
                              &first_access_teammate,
                              &first_access_opponent );

                long our_team_access_step = teammate_access_step;
                long opp_team_access_step = opponent_access_step;

#if ALLOW_SELF_PASS
                if ( self_access_step < our_team_access_step )
                {
                    our_team_access_step = self_access_step;
                }
#endif

                if ( unknown_access_step < opp_team_access_step )
                {
                    opp_team_access_step = unknown_access_step;
                }


                if ( max_diff < opp_team_access_step - our_team_access_step
                     && target_player->unum() == first_access_teammate )
                {
                    rcsc::Vector2D receive_point = rcsc::inertia_n_step_point
                                   ( result_pos,
                                     result_vel,
                                     our_team_access_step - kick_step,
                                     param.ballDecay() );

                    if( - param.pitchHalfLength() + 2.0 < receive_point.x
                     && receive_point.x < param.pitchHalfLength() - 2.0
                     && std::fabs( receive_point.y )
                        < param.pitchHalfWidth() - 2.0  
                     && receive_point.x >= wm.defenseLineX() + 1.0 )
                    {
                        target_point = receive_point;
                        max_diff = opp_team_access_step - our_team_access_step;
                        max_a = a;
                    }
                }
            }
#endif

#if defined(USE_SMART_KICK)
            rcsc::Body_SmartKick( target_point,
                                  target_ball_speed,
                                  param.ballSpeedMax() * 0.96,
                                  3 ).execute( agent );
#else
            rcsc::Body_KickMultiStep( target_point,
                                      target_ball_speed,
                                      true ).execute( agent );
#endif
            agent-> setNeckAction( new rcsc::Neck_TurnToPoint( target_player->pos() ) );

            agent->debugClient().setTarget( target_player->unum() );

            if ( max_diff != 0 )
            {
                agent->debugClient().addLine( wm.self().pos(), target_point );

                rcsc::dlog.addText( rcsc::Logger::ACTION,
                                    "max_a = %f", max_a );
                rcsc::dlog.addText( rcsc::Logger::ACTION,
                                    "max_diff = %ld", max_diff );
            }

            if ( agent->config().useCommunication()
                 && first_pass.player_number != rcsc::Unum_Unknown )
            {
                rcsc::dlog.addText( rcsc::Logger::ACTION,
                                    "%s:%d: execute() set pass communication."
                                    ,__FILE__, __LINE__ );
            }

            rcsc::Vector2D target_buf = target_point - agent->world().self().pos();
            target_buf.setLength( 1.0 );

            agent->addSayMessage( new rcsc::PassMessage( first_pass.player_number,
                                                         target_point + target_buf,
                                                         agent->effector().queuedNextBallPos(),
                                                         agent->effector().queuedNextBallVel() ) );
            return true;
            break;
        }

    case PassCourseTable::CourseSelection::Hold:
        //
        // clear ball
        //
        {
            const rcsc::Rect2D clear_area( param.ourTeamGoalLineX(),
                                           - param.penaltyAreaHalfWidth(),
                                           param.penaltyAreaLength() + 5.0,
                                           param.penaltyAreaWidth() );

            if ( clear_area.contains( wm.self().pos() )
                 || wm.existKickableOpponent() )
            {
                if ( Body_ClearBall2008().execute( agent ) )
                {
                    agent->setNeckAction( new rcsc::Neck_TurnToBallOrScan() );
                    agent->debugClient().addMessage( "clear ball" );
                    rcsc::dlog.addText( rcsc::Logger::TEAM, "clear ball" );
                    return true;
                }
            }
        }

#if 0
        //
        // addvance ball
        //
        {
            if ( wm.self().pos().x > wm.offsideLineX() - 30.0 )
            {
                Body_AdvanceBall2008().execute( agent );

                agent->debugClient().addMessage( "advance ball" );
                rcsc::dlog.addText( rcsc::Logger::TEAM, "advance ball" );
                return true;
                break;
            }
        }
#endif

        //
        // hold ball
        //
        {
            rcsc::Body_HoldBall().execute( agent );

            agent->debugClient().addMessage( "hold" );
            rcsc::dlog.addText( rcsc::Logger::TEAM, "hold" );
            return true;
            break;
        }
    }

#if 0
    rcsc::Vector2D pass_point;
    if ( rcsc::Body_Pass::get_best_pass( wm, &pass_point, NULL, NULL ) )
    {
        if ( pass_point.x > wm.self().pos().x - 1.0 )
        {
            bool safety = true;
            const rcsc::PlayerPtrCont::const_iterator opps_end = opps.end();
            for ( rcsc::PlayerPtrCont::const_iterator it = opps.begin();
                  it != opps_end;
                  ++it )
            {
                if ( (*it)->pos().dist( pass_point ) < 4.0 )
                {
                    safety = false;
                }
            }

            if ( safety )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: do best pass"
                                    ,__FILE__, __LINE__ );
                agent->debugClient().addMessage( "OffKickPass(1)" );
                rcsc::Body_Pass().execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                return true;
            }
        }
    }

    if ( nearest_opp_dist < 7.0 )
    {
        if ( rcsc::Body_Pass().execute( agent ) )
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: do best pass"
                                ,__FILE__, __LINE__ );
            agent->debugClient().addMessage( "OffKickPass(2)" );
            agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
            return true;
        }
    }

    // dribble to my body dir
    if ( nearest_opp_dist < 5.0
         && nearest_opp_dist > ( rcsc::ServerParam::i().tackleDist()
                                 + rcsc::ServerParam::i().defaultPlayerSpeedMax() * 1.5 )
         && wm.self().body().abs() < 70.0 )
    {
        const rcsc::Vector2D body_dir_drib_target
            = wm.self().pos()
            + rcsc::Vector2D::polar2vector(5.0, wm.self().body());

        int max_dir_count = 0;
        wm.dirRangeCount( wm.self().body(), 20.0, &max_dir_count, NULL, NULL );

        if ( body_dir_drib_target.x < rcsc::ServerParam::i().pitchHalfLength() - 1.0
             && body_dir_drib_target.absY() < rcsc::ServerParam::i().pitchHalfWidth() - 1.0
             && max_dir_count < 3
             )
        {
            // check opponents
            // 10m, +-30 degree
            const rcsc::Sector2D sector( wm.self().pos(),
                                         0.5, 10.0,
                                         wm.self().body() - 30.0,
                                         wm.self().body() + 30.0 );
            // opponent check with goalie
            if ( ! wm.existOpponentIn( sector, 10, true ) )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    "%s:%d: dribble to my body dir"
                                    ,__FILE__, __LINE__ );
                agent->debugClient().addMessage( "OffKickDrib(1)" );
                rcsc::Body_Dribble( body_dir_drib_target,
                                    1.0,
                                    rcsc::ServerParam::i().maxPower(),
                                    2
                                    ).execute( agent );
                agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
                return true;
            }
        }
    }

    rcsc::Vector2D drib_target( 50.0, wm.self().pos().absY() );
    if ( drib_target.y < 20.0 ) drib_target.y = 20.0;
    if ( drib_target.y > 29.0 ) drib_target.y = 27.0;
    if ( wm.self().pos().y < 0.0 ) drib_target.y *= -1.0;
    const rcsc::AngleDeg drib_angle = ( drib_target - wm.self().pos() ).th();

    // opponent is behind of me
    if ( nearest_opp_pos.x < wm.self().pos().x + 1.0 )
    {
        // check opponents
        // 15m, +-30 degree
        const rcsc::Sector2D sector( wm.self().pos(),
                                     0.5, 15.0,
                                     drib_angle - 30.0,
                                     drib_angle + 30.0 );
        // opponent check with goalie
        if ( ! wm.existOpponentIn( sector, 10, true ) )
        {
            const int max_dash_step
                = wm.self().playerType()
                .cyclesToReachDistance( wm.self().pos().dist( drib_target ) );
            if ( wm.self().pos().x > 35.0 )
            {
                drib_target.y *= ( 10.0 / drib_target.absY() );
            }

            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: fast dribble to (%.1f, %.1f) max_step=%d"
                                ,__FILE__, __LINE__,
                                drib_target.x, drib_target.y,
                                max_dash_step );
            agent->debugClient().addMessage( "OffKickDrib(2)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                std::min( 5, max_dash_step )
                                ).execute( agent );
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::TEAM,
                                "%s:%d: slow dribble to (%f, %f)"
                                ,__FILE__, __LINE__,
                                drib_target.x, drib_target.y );
            agent->debugClient().addMessage( "OffKickDrib(3)" );
            rcsc::Body_Dribble( drib_target,
                                1.0,
                                rcsc::ServerParam::i().maxPower(),
                                2
                                ).execute( agent );

        }
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return true;
    }

    // opp is far from me
    if ( nearest_opp_dist > 5.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: opp far. dribble(%f, %f)"
                            ,__FILE__, __LINE__,
                            drib_target.x, drib_target.y );
        agent->debugClient().addMessage( "OffKickDrib(4)" );
        rcsc::Body_Dribble( drib_target,
                            1.0,
                            rcsc::ServerParam::i().maxPower() * 0.4,
                            1
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return true;
    }

    // opp is near

    // can pass
    if ( rcsc::Body_Pass().execute( agent ) )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: pass"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "OffKickPass(3)" );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return true;
    }

    // opp is far from me
    if ( nearest_opp_dist > 3.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: opp far. dribble(%f, %f)"
                            ,__FILE__, __LINE__,
                            drib_target.x, drib_target.y );
        agent->debugClient().addMessage( "OffKickDrib(5)" );
        rcsc::Body_Dribble( drib_target,
                            1.0,
                            rcsc::ServerParam::i().maxPower() * 0.2,
                            1
                            ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return true;
    }

    if ( nearest_opp_dist > 2.5 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: hold"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "OffKickHold" );
        rcsc::Body_HoldBall().execute( agent );
        agent->setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );
        return true;
    }

    if ( wm.self().pos().x > wm.offsideLineX() - 10.0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: kick near side"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "OffKickToCorner" );
        Body_KickToCorner( (wm.self().pos().y < 0.0) ).execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
    }
    else
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "%s:%d: advance"
                            ,__FILE__, __LINE__ );
        agent->debugClient().addMessage( "OffKickAdvance" );
        rcsc::Body_AdvanceBall().execute( agent );
        agent->setNeckAction( new rcsc::Neck_ScanField() );
    }
#endif

    return true;
}


/*-------------------------------------------------------------------*/
/*!

*/
static
std::pair< bool, double >
evaluateSituation( const SimulatorUtil::SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent )
{
    const rcsc::ServerParam & param = rcsc::ServerParam::i();
    const rcsc::WorldModel & wm = agent->world();

    long our_team_access_step = std::min( s.teammate_access_step,
                                          s.self_access_step );
    long opp_team_access_step = std::min( s.opponent_access_step,
                                          s.unknown_access_step );

    rcsc::Vector2D receive_point = rcsc::inertia_n_step_point
                                         ( s.plan.getKickResultPos(),
                                           s.plan.getKickResultVel(),
                                           opp_team_access_step
                                           - s.plan.getKickStep(),
                                           param.ballDecay() );

    // if ball will be out of field, not a valid situation
    if ( receive_point.x < - param.pitchHalfLength() + 5.0
         || receive_point.x > + param.pitchHalfLength() - 5.0
         || std::fabs( receive_point.y ) > param.pitchHalfWidth() - 5.0 )
    {
        return std::make_pair( false, 0.0 );
    }

    if ( receive_point.x <= wm.offsideLineX() - 10.0 )
    {
        return std::make_pair( false, 0.0 );
    }

    if ( our_team_access_step >= opp_team_access_step )
    {
        // write 'X'
        const double d = 0.25;

        agent->debugClient().addLine
            ( receive_point + rcsc::Vector2D( -d, -d ),
              receive_point + rcsc::Vector2D( +d, +d ) );

        agent->debugClient().addLine
            ( receive_point + rcsc::Vector2D( -d, +d ),
              receive_point + rcsc::Vector2D( +d, -d ) );

        return std::make_pair( false, 0.0 );
    }
    else
    {
        // write 'O'
        agent->debugClient().addCircle( receive_point, 0.5 );
    }


    //
    // set basic evaluation
    //
    double value = receive_point.x;


#if 0
    //
    // subtract distance from goal
    //
    const double goal_dist = (receive_point - param.theirTeamGoalPos()).r();
    value -= goal_dist / 3.0;
#endif

#if 0
    //
    // add bonus if we get the ball
    //
    if ( our_team_access_step <= opp_team_access_step )
    {
        value += (opp_team_access_step - our_team_access_step + 1) * 100.0;
    }
#endif


#if 0
    //
    // add bonus by distance from receive_point to first access oppenent
    //
    if ( s.first_access_opponent != rcsc::Unum_Unknown )
    {
        value += (receive_point
                  - wm.opponent( s.first_access_opponent )->pos()).r();
    }
#endif

    return std::make_pair( true, value );
}

static
double
s_get_ball_speed_for_pass( double distance )
{
#if 1
    if ( distance >= 20.0 )
    {
        return 3.0;
    }
    else if ( distance >= 8.0 )
    {
        return 2.5;
    }
    else if ( distance >= 5.0 )
    {
        return 1.6;
    }
    else
    {
        return 1.5;
    }
#else
    return std::min( rcsc::calc_first_term_geom_series_last
                     (
#if 0
                      1.8,
#else
                      1.5,
#endif
                       distance,
                       rcsc::ServerParam::i().ballDecay() ),
                     rcsc::ServerParam::i().ballSpeedMax() );
#endif
}
