// -*-c++-*-

/*!
  \file body_get_ball.h
  \brief run to ball
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifndef BODY_GET_BALL_DEVEL_H
#define BODY_GET_BALL_DEVEL_H

#include <rcsc/player/soccer_action.h>
#include <rcsc/geom/vector_2d.h>
#include <rcsc/game_time.h>
#include <rcsc/player/world_model.h>
#include <rcsc/player/player_command.h>
#include <boost/shared_ptr.hpp>

/*!
  \class Body_GetBall
  \brief run to ball
*/
class Body_GetBall
    : public rcsc::BodyAction {

public:
    static
    double Default_Max_Error()
      {
          // negative value will be replaced by real default value
          // in initialize().
          return -1.0;
      }

private:
    //
    // cache
    //
    struct CacheInfo
    {
        rcsc::GameTime M_time_stamp;

        double M_max_dist;
        bool M_collide;
        bool M_use_back_dash;

        long M_gettable_step;
        rcsc::Vector2D M_gettable_point;
        boost::shared_ptr<rcsc::PlayerBodyCommand> M_command;

    public:
        CacheInfo()
            : M_time_stamp( -1, 0 ),
              M_max_dist( 0.0 ),
              M_collide( false ),
              M_use_back_dash( true ),
              M_gettable_step( 0 ),
              M_gettable_point( 0.0, 0.0 )
            , M_command( static_cast<rcsc::PlayerBodyCommand *>(0) )
          {
          }

        void calculate( const rcsc::WorldModel & wm ,
                        double max_dist,
                        bool collide,
                        bool use_back_dash );
    };

    static CacheInfo S_cache;

private:
    double M_max_dist;
    bool M_collide;
    bool M_use_back_dash;

public:
    /*!
      \brief constructor
      \param max_dist allowble distance, negative valiue means kickable area
      \param collide allow collidion or not
      \param use_back_dash allow back dash or not
     */
    Body_GetBall( double max_dist = -1.0,
                  bool collide = false,
                  bool use_back_dash = true );


    /*!
      \brief do go to ball
      \param agent agent pointer to the agent itself
      \return true with action, false if distance between ball and self
              less equals to max_dist
     */
    bool execute( rcsc::PlayerAgent * agent );


    static long gettable_step( const rcsc::WorldModel & wm,
                               double max_dist = Default_Max_Error(),
                               bool collide = false,
                               bool use_back_dash = true );

    static rcsc::Vector2D gettable_point( const rcsc::WorldModel & wm,
                                          double max_dist = Default_Max_Error(),
                                          bool collide = false,
                                          bool use_back_dash = true );

    static long spend_ball_time_for_ball_gettable
                        ( const rcsc::WorldModel & wm ,
                          long t,
                          double max_dist = Default_Max_Error(),
                          bool collide = false,
                          bool use_back_dash = true,
                          boost::shared_ptr<rcsc::PlayerBodyCommand> * M_command = 0);
};

#endif
