// -*-c++-*-

/*!
  \file neck_turn_to_ball_only.cpp
  \brief turn neck with attention to ball only
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "rcsc-ext/action/neck_turn_to_ball_only.h"
#include <rcsc/player/player_agent.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc/action/view_synch.h>
#include <rcsc/action/view_normal.h>
#include <rcsc/action/view_change_width.h>
#include <rcsc/math_util.h>

namespace rcsc_ext {

/*-------------------------------------------------------------------*/
/*!

*/
Neck_TurnToBallOnly::Neck_TurnToBallOnly()
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Neck_TurnToBallOnly::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__ ": Neck_TurnToBallOnly" );

    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    const rcsc::Vector2D next_ball_pos = agent->effector().queuedNextBallPos();
    const rcsc::Vector2D next_self_pos = agent->effector().queuedNextMyPos();
    const rcsc::AngleDeg next_self_body = agent->effector().queuedNextMyBody();
    const double next_view_half_width = agent->effector().queuedNextViewWidth() / 2.0;

    const rcsc::AngleDeg angle_diff = (next_ball_pos - next_self_pos).th()
                                      - next_self_body;

    const rcsc::SeeState & see_state = agent->seeState();
    const int see_cycles = see_state.cyclesTillNextSee() + 1;
    const bool can_see_next_cycle = (angle_diff.abs()
                                     < param.maxNeckAngle()
                                       + next_view_half_width);

    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        "see_cycles = %d",
                        see_cycles );
    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        "can_see_next_cycle = %s",
                        (can_see_next_cycle ? "true" : "false") );

    if ( can_see_next_cycle )
    {
        if ( see_cycles != 1 )
        {
            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                "set view to synch" );

#if 0
            agent->doChangeView( rcsc::ViewWidth::NARROW );
#elif 0
            rcsc::View_Synch().execute( agent );
#else
            rcsc::View_ChangeWidth( rcsc::ViewWidth::NARROW ).execute( agent );
#endif
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                "view untouched" );
        }
    }
    else
    {
        if ( see_cycles >= 2 )
        {
            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                "set view to narrow" );

#if 0
            agent->doChangeView( rcsc::ViewWidth::NARROW );
#elif 0
            rcsc::View_Synch().execute( agent );
#else
            rcsc::View_ChangeWidth( rcsc::ViewWidth::NARROW ).execute( agent );
#endif
        }
        else
        {
            rcsc::dlog.addText( rcsc::Logger::ACTION,
                                "set view to normal" );

#if 0
            agent->doChangeView( rcsc::ViewWidth::NORMAL );
#else
            rcsc::View_Normal().execute( agent );
#endif
        }
    }

    double target_angle = rcsc::bound( - param.maxNeckAngle(),
                                       angle_diff.degree(),
                                       + param.maxNeckAngle() );

    agent->doTurnNeck( target_angle - agent->world().self().neck() );

    return true;
}

Neck_TurnToBallOnly *
Neck_TurnToBallOnly::clone() const
{
    return new Neck_TurnToBallOnly();
}

}
