// -*-c++-*-

/*!
  \file simulator_util.cpp
  \brief convenience tools related simulator.
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "simulator_util.h"
#include "simulator.h"

#include <rcsc/player/world_model.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/common/server_param.h>

#include <rcsc/math_util.h>
using rcsc::EPS;

#include <utility>

/*-------------------------------------------------------------------*/
/*!

*/
boost::shared_ptr< SimulatorUtil::SimulatorEstimatedSituation >
SimulatorUtil::getBestChoice( std::pair< bool, double >
                              (*evaluator)( const SimulatorUtil::SimulatorEstimatedSituation & s,
                                            rcsc::PlayerAgent * agent ),
                              SimulatorUtil::KickPlanCont & candidates,
                              const rcsc::AbstractPlayerCont & players,
                              rcsc::PlayerAgent * agent,
                              bool self_add_kick_step,
                              int additional_self_penalty_step,
                              bool teammate_add_kick_step,
                              int additional_teammate_penalty_step,
                              bool opponent_add_kick_step,
                              int additional_opponent_penalty_step )
{
    const rcsc::WorldModel & wm = agent->world();

    Simulator sim;

    boost::shared_ptr< SimulatorUtil::SimulatorEstimatedSituation >
        best_situation
          ( static_cast< SimulatorUtil::SimulatorEstimatedSituation * >( 0 ) );

    double max_eval = 0.0;

    KickPlanCont::const_iterator c_end = candidates.end();
    for ( KickPlanCont::iterator it = candidates.begin();
          it != c_end;
          ++it )
    {
        const int kick_step = (*it).getKickStep();

        long self_access_step;
        long teammate_access_step;
        long opponent_access_step;
        long unknown_access_step;
        int first_access_teammate;
        int first_access_opponent;

        sim.simulate( wm,
                      players,
                      (*it).getKickResultPos(),
                      (*it).getKickResultVel(),
                      kick_step,
                      (self_add_kick_step? kick_step : 0)
                       + additional_self_penalty_step,
                      (teammate_add_kick_step? kick_step : 0)
                       + additional_teammate_penalty_step,
                      (opponent_add_kick_step? kick_step : 0)
                       + additional_opponent_penalty_step,
                      &self_access_step,
                      &teammate_access_step,
                      &opponent_access_step,
                      &unknown_access_step,
                      &first_access_teammate,
                      &first_access_opponent );

        boost::shared_ptr< SimulatorUtil::SimulatorEstimatedSituation >
            sit( new SimulatorEstimatedSituation
                                 ( (*it),
                                   self_access_step,
                                   teammate_access_step,
                                   opponent_access_step,
                                   unknown_access_step,
                                   first_access_teammate,
                                   first_access_opponent ) );

        std::pair< bool, double > eval = evaluator( *sit, agent );

        if ( eval.first )
        {
            if ( max_eval < eval.second
                 || ! best_situation )
            {
                max_eval = eval.second ;
                best_situation = sit;
            }
        }
    }

    return best_situation;
}
